// shaders.cpp
// IMPLEMENTACIA MANAGMENTU SHADEROV

#include "shaders.h"

// funkcia pre nacitanie textoveho suboru do pola znakov
GLchar* ReadShaderFile(const char *fileName)
{
    FILE* file = fopen(fileName, "rb" );
    if( file == NULL )
    {
        printf("Cannot open file %s. \n", fileName);
		return NULL;
    }

    // ziskaj velkost suboru v bajtoch
    fseek(file, 0L, SEEK_END);
    long fileSize = ftell(file);
    fseek(file, 0L, SEEK_SET);
    rewind(file);

    // nacitaj data zo suboru
    GLchar* buffer = new GLchar[fileSize + 1];
	long bytes = fread(buffer, 1, fileSize, file);
    buffer[bytes] = '\0';
	fclose(file);

	return buffer;
}

// nacitanie zdrojoveho suboru shadera a ulozenie do OpenGL shadera
// navratova hodnota je OpenGL id shadera
GLuint LoadShader(GLenum shaderType, const char *fileName)
{
    if (fileName == NULL)
        return 0;
    GLchar *shaderSource = ReadShaderFile(fileName);
    if (shaderSource == NULL)
        return 0;

    GLuint shader = glCreateShader(shaderType);
    glShaderSource(shader, 1, (const GLchar**)&shaderSource, NULL);
    glCompileShader(shader);
    delete shaderSource;

    GLint status;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
    if (status == GL_FALSE)  
    {
        GLchar log[1024];
        GLsizei length;
        glGetShaderInfoLog(shader, 1024, &length, log);
        printf("Failed compiling shader %s \n", fileName);
        printf("   Compilation error: %s \n", log);
        glDeleteShader(shader);

        return 0;
    }

    printf("Shader %s compiled successfully.\n", fileName);
    return shader;
}

// nacitanie sahderov a ich ulozenie do OpenGL
// navratova hodnota je OpenGL id shader programu
GLuint LoadProgram(const char *fileVertexShader, const char *fileGeometryShader, const char *fileFragmentShader)
{
    // nacitanie a kompilacia shaderov
    GLuint vertexShader = LoadShader(GL_VERTEX_SHADER, fileVertexShader);
    GLuint geometryShader = LoadShader(GL_GEOMETRY_SHADER, fileGeometryShader);
    GLuint fragmentShader = LoadShader(GL_FRAGMENT_SHADER, fileFragmentShader);

    // vytvorenie program objektu a pripojenie shaderov
    GLuint programObj = glCreateProgram();
    if (vertexShader > 0)
        glAttachShader(programObj, vertexShader);
    if (geometryShader > 0)
        glAttachShader(programObj, geometryShader);
    if (fragmentShader > 0)
        glAttachShader(programObj, fragmentShader);

    // pridaj vlastnosti geometry shadra, ak ho mame
    if (geometryShader > 0 && glProgramParameteriARB != NULL)
    {
        glProgramParameteriARB(programObj, GL_GEOMETRY_INPUT_TYPE_ARB, GL_POINTS);
        glProgramParameteriARB(programObj, GL_GEOMETRY_OUTPUT_TYPE_ARB, GL_TRIANGLE_STRIP);
        int temp;
        glGetIntegerv(GL_MAX_GEOMETRY_OUTPUT_VERTICES_ARB, &temp);
        glProgramParameteriARB(programObj, GL_GEOMETRY_VERTICES_OUT_ARB,temp);
    }
    glLinkProgram(programObj);

    // zisti uspesnost zlinkovania shaderov
    GLint status;
    glGetProgramiv(programObj, GL_LINK_STATUS, &status);
    if (status == GL_FALSE)  
    {
        GLchar log[1024];
        GLsizei length;
        glGetProgramInfoLog(programObj, 1024, &length, log);
        printf("Failed linking shaders with error %s\n", log);

        glDeleteProgram(programObj);
        glDeleteShader(fragmentShader);
        glDeleteShader(geometryShader);
        glDeleteShader(vertexShader);

        return 0;
    }

    printf("Shaders linked successfully.\n");
    return programObj;
}

void ClearProgram(GLuint program)
{
    // ziskaj vsetky shader objekty pripojene ku programu
    GLuint shaders[255];
    GLsizei count = 0;
    glGetAttachedShaders(program, 255, &count, shaders);

    for (int i = 0; i < count; i++)
        glDeleteShader(shaders[i]);
    glDeleteProgram(program);
}

// setting uniform variable inside OpenGL shader program
bool SetShaderUniform1i(GLuint program, const char* name, GLint value)
{
    int location = glGetUniformLocation(program, name);
    if (location == -1)
    {
        printf("Error getting location for uniform %s.\n", name);
        return false;
    }

    glUniform1i(location, value);
    return true;
}

bool SetShaderUniform1f(GLuint program, const char* name, GLfloat value)
{
    int location = glGetUniformLocation(program, name);
    if (location == -1)
    {
        printf("Error getting location for uniform %s.\n", name);
        return false;
    }

    glUniform1f(location, value);
    return true;
}

bool SetShaderUniform2f(GLuint program, const char* name, GLfloat* values)
{
    int location = glGetUniformLocation(program, name);
    if (location == -1)
    {
        printf("Error getting location for uniform %s.\n", name);
        return false;
    }

    glUniform2fv(location, 1, values);
    return true;
}

bool SetShaderUniform3f(GLuint program, const char* name, GLfloat* values)
{
    int location = glGetUniformLocation(program, name);
    if (location == -1)
    {
        printf("Error getting location for uniform %s.\n", name);
        return false;
    }

    glUniform3fv(location, 1, values);
    return true;
}

bool SetShaderUniform4f(GLuint program, const char* name, GLfloat* values)
{
    int location = glGetUniformLocation(program, name);
    if (location == -1)
    {
        printf("Error getting location for uniform %s.\n", name);
        return false;
    }

    glUniform4fv(location, 1, values);
    return true;
}

bool SetShaderUniformMatrix4f(GLuint program, const char* name, GLfloat* values)
{
    int location = glGetUniformLocation(program, name);
    if (location == -1)
    {
        printf("Error getting location for uniform %s.\n", name);
        return false;
    }

    glUniformMatrix4fv(location, 1, GL_FALSE, values);
    return true;
}